module Constant_Queue(Queue, empty, isEmpty, dequeue, enqueue) 
  where

-- Queue implementation with constant time operations
-- Algorithm due to Chris Okasaki.

-- invariant: 
-- length xs = length ys + length zs
-- zs is a suffix of xs, in fact they share a certain evaluated part
-- (xs,ys,zs) represents queue xs ++ reverse ys

newtype Queue a = Queue ([a], [a], [a]) 

empty :: Queue a
empty = Queue ([], [], [])

isEmpty :: Queue a -> Bool
isEmpty (Queue (xs, _, _)) = null xs

queue = Queue . normalize

enqueue :: a -> Queue a -> Queue a
enqueue x (Queue (xs, ys, zs)) = queue (xs, x : ys, zs)

dequeue :: Queue a -> (a, Queue a)
dequeue (Queue ([], _, [])) = error "dequeue on empty queue"
dequeue (Queue (x : xs, ys, zs)) = (x, queue (xs, ys, zs))

normalize (xs, ys, []) = (zs, [], zs) 
  where zs = rot xs ys []
normalize (xs, ys, z : zs) = (xs, ys, zs)

-- rot xs ys zs = xs ++ reverse ys ++ zs
-- under condition length xs + 1 >= length ys
-- first list element is produced in one step
rot (x : xs) [] zs = x : rot xs [] zs
rot (x : xs) (y : ys) zs = x : rot xs ys (y : zs)
rot [] [y] zs = y : zs
rot [] [] zs = zs
